package lightcontainer.domains.client;

import lightcontainer.domains.server.UnicastServerListener;
import lightcontainer.interfaces.ClientHandlerFFE;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.interfaces.UnicastCHR;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;
import lightcontainer.protocol.rules.reader.SigninRule;
import lightcontainer.protocol.rules.reader.SignoutRule;
import lightcontainer.protocol.rules.reader.SignupRule;
import lightcontainer.protocol.rules.writer.SignErrorRule;
import lightcontainer.protocol.rules.writer.SignOkRule;

import javax.crypto.BadPaddingException;
import javax.net.ssl.SSLHandshakeException;
import java.io.*;
import java.net.Socket;
import java.nio.charset.StandardCharsets;

/**
 * ClientHandler
 *
 * <!> UNICAST CLIENT <!>
 * Class communicating with the client, and
 * intercepting and sending files to the client.
 *
 * @version     1.1
 * @since       1.0
 *
 * @see         Runnable
 * @see         AutoCloseable
 * @author      Jérémi NIHART <j.nihart@student.helmo.be>
 */
public class ClientHandler implements Runnable, AutoCloseable {
    // Variables
    private ClientHandlerFFE fileFrontEnd;
    private final Socket client;
    private ProtocolRepository protocolRep;
    private Context context;
    private boolean client_run;
    private UnicastCHR repository;


    private BufferedReader reader;
    private PrintWriter writer;
    private ProtocolWriter.ProtocolResult response;

    // Constructor
    public ClientHandler(UnicastCHR repository, Socket client, ClientHandlerFFE ffe, ProtocolRepository protocolRep, Context context) {
        this.repository = repository;
        this.fileFrontEnd = ffe;
        this.client = client;
        this.protocolRep = protocolRep;
        this.context = context;
        this.client_run = false;
        initClient();
    }

    /**
     * Initialise the Client's Reader and Writer.
     *
     * @since       1.0
     *
     * @see         BufferedReader
     * @see         PrintWriter
     */
    private void initClient() {
        // Start the thread
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                this.client.getInputStream(),
                StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                this.client.getOutputStream(),
                StandardCharsets.UTF_8
            ), true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread Function
     * Start the dialogue with the client.
     *
     * @since       1.0
     */
    @Override
    public void run() {
        this.client_run = true;
        while (this.client_run) {
            // Signifie le démarrage d'une nouvelle rquête
            context.newBundle();

            try {
                String command = this.reader.readLine();
                if (command != null) {
                    System.out.println("Client: " + command);
                } else {
                    repository.disconnect(this);
                    break;
                }

                ProtocolReader.ProtocolResult ruleResult = protocolRep.executeReader(context, command + "\r\n");
                if (ruleResult == null) {
                    repository.disconnect(this);
                    break;
                }

                if (checkAccess(ruleResult)) {
                    // Lecture du fichier client
                    ruleResult.read(this.client.getInputStream());
                    System.out.println(context.getLogin() + " - " + 1);
                    ProtocolWriter.ProtocolResult writerCommand = ruleResult.getResultCommand();
                    // TODO : Vérifier que le StorBackEnd demandé (Pas toujours demandé) est disponible
                    if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND && !fileFrontEnd.canExecuteCommand(ruleResult.getRequestDomain())) {
                        System.out.println("ICI");
                        System.out.println(context.getLogin() + " - " + 2);
                        writer.print(ruleResult.onNotExecutable(context)); // Renvoie au client
                        writer.flush();
                    } else if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND) {

                        fileFrontEnd.newCommand(context, writerCommand); // Envoie dans la file de tâche FileFrontEnd en attente d'un traitement d'un StorBackEnd

                        System.out.println(context.getLogin() + " - " + 3);
                        // Attend la fin de la réalisation de la tâche
                        waitTaskResponse();
                        System.out.println(context.getLogin() + " - " + 4);

                        writer.write(response.getCommand()); // Renvoie au client
                        writer.flush();
                        response.write(this.client.getOutputStream()); // Ecrit au client si nécessaire
                        System.out.println(context.getLogin() + " - " + 5);
                    } else {
                        System.out.println(context.getLogin() + " - " + 6);
                        writer.print(writerCommand.getCommand()); // Renvoie au client
                        writer.flush();
                    }

                } else {
                    System.out.println(4);
                    accessDenied();
                }
            } catch (IOException e) {
                repository.disconnect(this);
                break;
            }
        }

        try {
            this.reader.close();
            this.writer.close();
            this.client.close();
            System.out.printf("[CLIENT] %s s'est déconnecté\n", context.getLogin());
        } catch (IOException ignored) { }
    }

    /**
     * Permet de vérifier si le client possède l'accès demandé
     * @param ruleResult La règle
     * @return TRUE si le client possède l'accès demandé
     */
    private boolean checkAccess(ProtocolReader.ProtocolResult ruleResult) {
        checkSignout(ruleResult);
        if (context.isConnected())
            return true;

        try {
            ruleResult
                    .getClass()
                    .asSubclass(SignupRule.Result.class);
            return true;
        } catch (ClassCastException e) {
            try {
                ruleResult.getClass().asSubclass(SigninRule.Result.class);
                return true;
            } catch (ClassCastException e2) { }
        }

        return false;
    }

    /**
     * Lorsque l'accès à été refusé.
     */
    private void accessDenied() {
        ProtocolWriter.ProtocolResult signErrorResult = protocolRep.executeWriter(context, SignErrorRule.NAME);
        writer.write(signErrorResult.getCommand()); // Envoie SignError car echec de la connection
        writer.flush();
    }

    /**
     * Vérifie s'il s'âgit d'une demande de déconnexion
     * @param ruleResult
     */
    private void checkSignout(ProtocolReader.ProtocolResult ruleResult) {
        try {
            ruleResult.getClass().asSubclass(SignoutRule.Result.class);
            repository.disconnect(this);
        } catch (ClassCastException e2) { }
    }

    /**
     * Vérifie s'il s'âgit d'une demande de déconnexion
     * @param ruleResult
    private void checkSignError(ProtocolWriter.ProtocolResult ruleResult) {
       if (ruleResult.getCommand().startsWith(SignErrorRule.NAME)) {
            System.out.println("Pas pu connecter");
            repository.disconnect(this);
        }
    }
     */

    /**
     * Permet au Client d'attendre la fin de la réalisation de sa tâche
     */
    private void waitTaskResponse() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) { e.printStackTrace(); }
        }
    }

    /**
     * Permet d'envoyer la réponse au client.
     * @param response La réponse
     */
    public void respond(ProtocolWriter.ProtocolResult response) {
        this.response = response;
        synchronized (this) {
            this.notify();
        }
    }

    /**
     * AutoClosable Function
     * Close the Client thread and resources.
     *
     * @since       1.0
     */
    @Override
    public void close() {
        System.out.println("Call close");
        this.client_run = false;
    }

    public String getLogin() {
        return this.context.getLogin();
    }

}
